'use strict';

var debug = require('debug')('metalsmith:transclude');
var hercule = require('hercule');
var async = require('async');
var path = require('path');
var match = require('multimatch');
var stream = require('stream');
var pointer = require('json-pointer');
var omit = require('lodash/omit');

/**
 * Expose `plugin`.
 */

module.exports = plugin;

/**
 * Metalsmith plugin to transclude content.
 *
 *
 * @param {Object} options
 * @param {string} options.frontmatter Include frontmatter in parent file.
 *
 * @return {Function}
 */

// It feels like a better separation of concerns (particularly to help troubleshoot)
// to have the frontmatter included as yaml in the parent file's content rather than in
// its metadata key. Maybe this could be an option (but I'm not sure how the recursive hercule
// resolution would deal with nested keys whereas the current approach should be more predictable)

function plugin(options) {
  var _ref = options || {},
      _ref$patterns = _ref.patterns,
      patterns = _ref$patterns === undefined ? ['**/*.md'] : _ref$patterns,
      _ref$comments = _ref.comments,
      comments = _ref$comments === undefined ? false : _ref$comments,
      _ref$frontmatter = _ref.frontmatter,
      frontmatter = _ref$frontmatter === undefined ? false : _ref$frontmatter,
      _ref$verbose = _ref.verbose,
      verbose = _ref$verbose === undefined ? true : _ref$verbose,
      _ref$warning = _ref.warning,
      warning = _ref$warning === undefined ? true : _ref$warning;

  return function transclude(files, metalsmith, done) {
    var processedFiles = {};
    async.eachOfSeries(files, function (file, key, cb) {
      debug('>> Processing %s', key);

      if (match(key, patterns).length === 0) {
        debug('skip', key);
        return cb(); // do nothing
      }

      // TODO: Submit issue to hercule to allow returning context from resolver to avoid mutable state.
      var current_metadata = {};

      function resolveMetalsmith(url, sourcePath) {
        debug('>>> resolveMetalsmith             :', url);
        sourcePath !== 'string' && debug('>>> SourcePath                    :', sourcePath);

        var isLocalUrl = /^[^ ()"']+/;
        if (!isLocalUrl.test(url)) return null;

        var isSectionUrl = /#+/;
        var section = url.match(isSectionUrl);

        var baseUrl = section ? url.slice(0, section.index) : url;

        // If there is a sourcePath then we're in a nested transclusion

        // const relativePath = path.dirname(sourcePath);
        // const targetKey = path.join(path.dirname(key), url);
        var targetKey = sourcePath !== 'string' ? path.join(path.dirname(sourcePath), baseUrl) : path.join(path.dirname(key), baseUrl);

        debug('>>> Using target key              :', targetKey);
        var resolvedKey = files[targetKey] && targetKey || files[targetKey + '.md'] && targetKey + '.md';
        if (!resolvedKey) {
          if (warning) console.log(`Missing transclusion destination in ${key}: ${targetKey}`);
          return null;
        }
        debug('>>> Found target file             :', resolvedKey);

        // TODO: This should merge the frontmatter with the pipeline file metadata with
        // a priority to the file metadata to avoid suprises
        // if this plugin is after file metadata modification steps
        //
        // debug('section', section);

        var transcluded = section ? extractSection(files[resolvedKey].contents.toString(), section.input.slice(section.index + section[0].length), section[0]) : files[resolvedKey].contents.toString();
        var metadata = omit(files[resolvedKey], ['contents', 'mode', 'stats']);

        debug('Processes frontmatter. metadata', metadata);
        debug('Processes frontmatter. transcluded', '\n' + transcluded);

        // Local mutation, merge with current transcluded file metadata object.
        // current_metadata = { [url]: metadata };
        frontmatter && pointer.set(current_metadata, '/' + baseUrl.replace('.md', ''), metadata);

        var content = new stream.Readable({ encoding: 'utf8' });
        if (comments) content.push(`<!-- Following snippet transcluded from ${resolvedKey} ${verbose ? 'with resolveMetalsmith.url(' + baseUrl + ')' : ''} ${verbose ? JSON.stringify(metadata) : ''} -->\n`);
        content.push(transcluded);
        if (comments) content.push(`\n<!-- End of transcluded snippet from ${resolvedKey} -->\n\n`);
        content.push(null);

        return {
          content,
          url: resolvedKey
        };
      }

      var resolvers = [resolveMetalsmith,
      // resolveMetalsmithPattern,
      function (url, source, placeholder) {
        return { content: placeholder };
      }];
      // console.time('hercule')
      hercule.transcludeString(file.contents, { resolvers }, function (err, result) {
        // console.timeEnd('hercule')
        // if (err && err.code === 'ENOENT') {
        //   debug("Couldn't find the following file and skipped it. " + err.path);
        //   return cb();
        // }
        if (err) {
          console.error(err);
          return cb(err);
        }
        // mutate global files array.
        debug('<< Finished processing file: ', key);
        processedFiles[key] = {};
        if (frontmatter) processedFiles[key].metadata = current_metadata;
        if (result) processedFiles[key].contents = result;
        return cb();
      });
    }, function (err) {
      if (err) return done(err);
      Object.keys(processedFiles).forEach(function (key) {
        // Add frontmatter as metadata as well.
        files[key] = Object.assign(files[key], processedFiles[key].metadata, {
          contents: processedFiles[key].contents
        });
      });

      debug('Transcluded!');
      done();
    });
    var extractSection = function extractSection(source, section, mode) {
      // debug('source', source);
      // debug('section', section);
      // debug('mode', mode);
      var regexp = new RegExp('^#+ +' + section.replace(/-/g, '[\\W_]'), 'im');
      // First find the matching starting line for the section header
      var match = source.match(regexp);
      var start = match.index;
      // Depending on the mode, find the ending line
      if (mode === '#') {
        // In this mode include only until the next header.
        var endMatch = source.slice(start + match[0].length).match(/^#+ +(.*)$/m);
        return endMatch ? source.slice(start, start + endMatch.index + match[0].length) : source.slice(start);
      } else if (mode === '##') {
        // In this mode include until before the next header at the same level (or until the end of the file)
        var sameLevel = new RegExp('^' + match[0].match(/^#+/)[0] + ' +(.*)$', 'm');
        var _endMatch = source.slice(start + match[0].length).match(sameLevel);
        return _endMatch ? source.slice(start, start + _endMatch.index + match[0].length) : source.slice(start);
      }
    };
  };
}